'use strict';

var g = ohm.grammarFromScriptElement();
var F = new Language(g, g.semantics().addOperation('toAST', {

  Exp_let: function(_let, p, _eq, e1, _in, e2) {
    return new Let(p.toAST(), e1.toAST(), e2.toAST());
  },

  Exp_letrec: function(_let, _rec, x, _eq, e1, _in, e2) {
    return new LetRec(x.toAST(), e1.toAST(), e2.toAST());
  },

  Exp_typedLetrec: function(_let, _rec, _op, x, _colon, t, _cp, _eq, e1, _in, e2) {
    return new TypedLetRec(x.toAST(), t.toAST(), e1.toAST(), e2.toAST());
  },

  Exp_fun: function(_fun, ps, _arr, e) {
    return ps.toAST().reverse().reduce(
        (body, p) => new Fun(p, body),
        e.toAST());
  },

  Exp_if: function(_if, cond, _then, tb, _else, fb) {
    return new If(cond.toAST(), tb.toAST(), fb.toAST());
  },

  Exp_match: function(_match, e, _with, _optBar, pes) {
    var pesAST = pes.toAST();
    return new Match(
        e.toAST(),
        pesAST.map(function(pe) { return pe.p; }),
        pesAST.map(function(pe) { return pe.e; }));
  },

  PatAndExp: function(p, _arr, e) {
    return {p: p.toAST(), e: e.toAST()};
  },

  Pat_datum: function(C, p) {
    return new Datum(C.toAST(), p.toAST());
  },

  PriPat_emptyDatum: function(C) {
    return new Datum(C.toAST(), new Tuple([]));
  },

  PriPat_paren: function(_op, p, _cp) {
    return p.toAST();
  },

  PriPat_typed: function(_op, p, _colon, t, _cp) {
    return new Typed(p.toAST(), t.toAST());
  },

  PriPat_tuple: function(_op, ps, _cp) {
    return new Tuple(ps.toAST());
  },

  PriPat_list: function(_os, ps, _cs) {
    return ps.toAST().reverse().reduce(
        function(lst, p) { return new Datum('Cons', new Tuple([p, lst])); },
        new Datum('Nil', new Tuple([])));
  },

  PriPat_wild: function(x) {
    return new Wildcard();
  },

  PriPat_ident: function(x) {
    return new Var(x.toAST());
  },

  PriPat_number: function(n) {
    return new Lit(n.toAST());
  },

  PriPat_true: function(_true) {
    return new Lit(true);
  },

  PriPat_false: function(_false) {
    return new Lit(false);
  },

  OrExp_or: function(x, _or, y) {
    return new BinOp('||', x.toAST(), y.toAST());
  },

  AndExp_and: function(x, _and, y) {
    return new BinOp('&&', x.toAST(), y.toAST());
  },

  EqExp_eq: function(x, _eq, y) {
    return new BinOp('=',  x.toAST(), y.toAST());
  },

  EqExp_neq: function(x, _neq, y) {
    return new BinOp('!=', x.toAST(), y.toAST());
  },

  RelExp_lt: function(x, _lt, y) {
    return new BinOp('<', x.toAST(), y.toAST());
  },

  RelExp_gt: function(x, _gt, y) {
    return new BinOp('>', x.toAST(), y.toAST());
  },

  AddExp_plus: function(x, _plus, y) {
    return new BinOp('+', x.toAST(), y.toAST());
  },

  AddExp_minus: function(x, _minus, y) {
    return new BinOp('-', x.toAST(), y.toAST());
  },

  MulExp_times: function(x, _times, y) {
    return new BinOp('*', x.toAST(), y.toAST());
  },

  MulExp_divide: function(x, _div, y) {
    return new BinOp('/', x.toAST(), y.toAST());
  },

  MulExp_modulus: function(x, _mod, y) {
    return new BinOp('%', x.toAST(), y.toAST());
  },

  CallExp_call: function(f, e) {
    return new Call(f.toAST(), e.toAST());
  },

  UnExp_pos: function(_plus, e) {
    return e.toAST();
  },

  UnExp_neg: function(_minus, e) {
    return new BinOp('-', new Lit(0), e.toAST());
  },

  UnExp_delay: function(_delay, e) {
    return new Delay(e.toAST());
  },

  UnExp_force: function(_force, e) {
    return new Force(e.toAST());
  },

  DatumExp_datum: function(C, e) {
    return new Datum(C.toAST(), e.toAST());
  },

  PriExp_emptyDatum: function(C) {
    return new Datum(C.toAST(), new Tuple([]));
  },

  PriExp_paren: function(_op, e, _cp) {
    return e.toAST();
  },

  PriExp_typed: function(_op, e, _colon, t, _cp) {
    return new Typed(e.toAST(), t.toAST());
  },

  PriExp_tuple: function(_op, es, _cp) {
    return new Tuple(es.toAST());
  },

  PriExp_listComp: function(_os, e, _bar, p, _arr, el, _optComma, optEp, _cs) {
    return new ListComp(
        e.toAST(),
        p.toAST(),
        el.toAST(),
        optEp.toAST()[0]);
  },

  PriExp_list: function(_os, es, _cs) {
    return es.toAST().reverse().reduce(
        function(lst, e) { return new Datum('Cons', new Tuple([e, lst])); },
        new Datum('Nil', new Tuple([])));
  },

  PriExp_ident: function(x) {
    return new Var(x.toAST());
  },

  PriExp_number: function(n) {
    return new Lit(n.toAST());
  },

  PriExp_true: function(_true) {
    return new Lit(true);
  },

  PriExp_false: function(_false) {
    return new Lit(false);
  },

  FunType_fun: function(ta, _arr, tr) {
    return new FunType(ta.toAST(), tr.toAST());
  },

  TupleType_tuple: function(t, _timeses, ts) {
    return new TupleType([t.toAST()].concat(ts.toAST()));
  },

  ListOrDelayedType_list: function(t, _list) {
    return new ListType(t.toAST());
  },

  ListOrDelayedType_delayed: function(t, _delayed) {
    return new DelayedType(t.toAST());
  },

  PriType_paren: function(_op, t, _cp) {
    return t.toAST();
  },

  PriType_int: function(_) {
    return new IntType();
  },

  PriType_bool: function(_) {
    return new BoolType();
  },

  PriType_unit: function(_) {
    return new TupleType([]);
  },

  typeVar: function(_quote, x) {
    return new TypeVar(x.toAST());
  },

  ident: function(_first, _rest) {
    return this.interval.contents;
  },

  ctor: function(_first, _rest) {
    return this.interval.contents;
  },

  number: function(_) {
    return parseFloat(this.interval.contents);
  },

  NonemptyListOf: function(x, _sep, xs) {
    return [x.toAST()].concat(xs.toAST());
  },

  EmptyListOf: function() {
    return [];
  }

}));

// F.prettyPrintAST and F.prettyPrintValue are declared in prettyPrint.js
