'use strict';

tests(
  F,
  {
    name: 'recursive let',
    code: 'let rec f = fun n -> if n = 0\n' +
          '                     then 1\n' +
          '                     else n * f (n - 1) in\n' +
          '  f 5',
    expected: 120
  },
  {
    name: 'tuples',
    code: '(1, 2, 3 + 4)',
    expected: new Tuple([1, 2, 7])
  },
  {
    name: 'data constructors',
    code: 'C (5 + 4, D (false, E))',
    expected: new Datum(
                'C',
                new Tuple([
                  9,
                  new Datum(
                    'D',
                    new Tuple([
                      false,
                      new Datum('E', new Tuple([]))]))]))
  },
  {
    name: 'list sugar (should work if you got data constructors working)',
    code: '[5 + 4;0;6]',
    expected: new Datum('Cons', new Tuple([
                9,
                new Datum('Cons', new Tuple([
                  0,
                  new Datum('Cons', new Tuple([
                    6,
                    new Datum('Nil', new Tuple([]))]))]))]))
  },
  {
    name: 'match',
    code: 'let lst = [1; Pair (2, 3); 4] in\n' +
          '  match lst with\n' +
          '    [1; Pair (x, 3); y] -> y * 10 + x',
    expected: 42
  },
  {
    name: 'match failure should throw exception',
    code: 'match 5 with 6 -> 42',
    shouldThrow: true
  },
  {
    name: 'factorial w/ pattern matching',
    code: 'let rec f = fun n ->\n' +
          '              match n with\n' +
          '                0 -> 1\n' +
          '              | _ -> n * f (n - 1) in\n' +
          '  f 6',
    expected: 720
  },
  {
    name: 'map',
    code: 'let rec map = fun f l ->\n' +
          '                match l with\n' +
          '                  Nil -> Nil\n' +
          '                | Cons (x, xs) -> Cons (f x, map f xs) in\n' +
          '  map (fun x -> x + 1) [1;2;3]',
    expected: new Datum('Cons', new Tuple([
                2,
                new Datum('Cons', new Tuple([
                  3, new Datum('Cons', new Tuple([4,
                    new Datum('Nil', new Tuple([]))]))]))]))
  },
  {
    name: 'destructuring let',
    code: 'let (x, y) = (1, 2 + 3) in x + y',
    expected: 6
  },
  {
    name: 'destructuring let (match failure)',
    code: 'let (Cons (x, Cons (y, Nil))) = Cons (1 + 2, Foo (7, Nil)) in x + y',
    shouldThrow: true
  },
  {
    name: 'destructuring fun',
    code: '(fun (Point (x, y)) -> x - y) (Point (3, 4))',
    expected: -1
  },
  {
    name: 'destructuring fun (match failure()',
    code: '(fun 5 -> true) 4',
    shouldThrow: true
  },
  {
    name: 'list comprehension w/o predicate',
    code: 'let nats = [0;1;2;3;4] in\n' +
          '  [x * 2 | x <- nats]',
    expected: new Datum('Cons', new Tuple([
                0,
                new Datum('Cons', new Tuple([
                  2, new Datum('Cons', new Tuple([4,
                    new Datum('Cons', new Tuple([6,
                      new Datum('Cons', new Tuple([8,
                        new Datum('Nil', new Tuple([]))]))]))]))]))]))
  },
  {
    name: 'list comprehension w/ predicate',
    code: 'let nats = [0;1;2;3;4] in\n' +
          '  [x * 2 | x <- nats, x % 2 = 0]',
    expected: new Datum('Cons', new Tuple([0,
                new Datum('Cons', new Tuple([4,
                  new Datum('Cons', new Tuple([8,
                    new Datum('Nil', new Tuple([]))]))]))]))
  },
  {
    name: 'delay and force',
    code: 'let rec take = fun n s ->\n' +
          '  match n with\n' +
          '    0 -> Nil\n' +
          '  | _ -> match s with\n' +
          '           Cons (first, rest) ->\n' +
          '             Cons (first, take (n - 1) (force rest)) in\n' +
          'let rec ones = Cons (1, delay ones) in\n' +
          '  take 5 ones',
    expected:  new Datum('Cons', new Tuple([1,
                 new Datum('Cons', new Tuple([1,
                   new Datum('Cons', new Tuple([1,
                     new Datum('Cons', new Tuple([1,
                       new Datum('Cons', new Tuple([1,
                         new Datum('Nil', new Tuple([]))]))]))]))]))]))
  }
);

