'use strict';

// ---------------------------------------------------------
// Classes that represent AST nodes
// ---------------------------------------------------------

// Classes for Assignment 5

class Lit extends AST {
  constructor(primValue) {
    super();
    this.primValue = primValue;
  }
}

class BinOp extends AST {
  constructor(op, e1, e2) {
    super();
    this.op = op;
    this.e1 = e1;
    this.e2 = e2;
  }
}

class If extends AST {
  constructor(e1, e2, e3) {
    super();
    this.e1 = e1;
    this.e2 = e2;
    this.e3 = e3;
  }
}

class Var extends AST {
  constructor(x) {
    super();
    this.x = x;
  }
}

class Let extends AST {
  constructor(p, e1, e2) {
    super();
    this.p = p;
    this.e1 = e1;
    this.e2 = e2;
  }
}

class Fun extends AST {
  constructor(p, e) {
    super();
    this.p = p;
    this.e = e;
  }
}

class Call extends AST {
  constructor(ef, ea) {
    super();
    this.ef = ef;
    this.ea = ea;
  }
}

class Closure {
  constructor(fun, env) {
    this.fun = fun;
    this.env = env;
  }
}


// Classes for Assignment 6

class LetRec extends AST {
  constructor(x, e1, e2) {
    super();
    this.x = x;
    this.e1 = e1;
    this.e2 = e2;
  }
}

class TypedLetRec extends AST {
  constructor(x, t, e1, e2) {
    super();
    this.x = x;
    this.t = t;
    this.e1 = e1;
    this.e2 = e2;
  }
}

class Tuple extends AST {
  constructor(es) {
    super();
    this.es = es;
  }

  get vs() {
    return this.es;
  }

  get ps() {
    return this.es;
  }
}

class Datum extends AST {
  constructor(C, e) {
    super();
    this.C = C;
    this.e = e;
  }

  get v() {
    return this.e;
  }

  get p() {
    return this.e;
  }
}

class Match extends AST {
  constructor(e, ps, es) {
    super();
    this.e = e;
    this.ps = ps;
    this.es = es;
  }
}

class Wildcard extends AST {
  constructor() {
    super();
  }
}

class ListComp extends AST {
  constructor(e, p, elist, epred) {
    super();
    this.e = e;
    this.p = p;
    this.elist = elist;
    this.epred = epred;
  }
}

class Delay extends AST {
  constructor(e) {
    super();
    this.e = e;
  }
}

class Force extends AST {
  constructor(e) {
    super();
    this.e = e;
  }
}

class Typed extends AST {
  constructor(e, t) {
    super();
    this.e = e;
    this.t = t;
  }

  get p() {
    return this.e;
  }

  get v() {
    return this.e;
  }
}

class Type extends AST {
  constructor() {
    super();
  }
}

class FunType extends Type {
  constructor(ta, tr) {
    super();
    this.ta = ta;
    this.tr = tr;
  }
}

class TupleType extends Type {
  constructor(ts) {
    super();
    this.ts = ts;
  }
}

class ListType extends Type {
  constructor(t) {
    super();
    this.t = t;
  }
}

class DelayedType extends Type {
  constructor(t) {
    super();
    this.t = t;
  }
}

class IntType extends Type {
  constructor() {
    super();
  }
}

class BoolType extends Type {
  constructor() {
    super();
  }
}

class TypeVar extends Type {
  constructor(x) {
    super();
    this.x = x;
  }
}

// a type to denote an unannotated Nil value
class NilType extends Type {
  constructor() {
    super();
  }
}

